@@ -10,7 +10,7 @@ typedef struct _IBusM17NEngine IBusM17NEngine;
 typedef struct _IBusM17NEngineClass IBusM17NEngineClass;
 
 struct _IBusM17NEngine {
-	IBusEngine parent;
+    IBusEngine parent;
 
     /* members */
     MInputContext *context;
@@ -20,22 +20,23 @@ struct _IBusM17NEngine {
 };
 
 struct _IBusM17NEngineClass {
-	IBusEngineClass parent;
+    IBusEngineClass parent;
 };
 
 /* functions prototype */
-static void	ibus_m17n_engine_class_init	    (IBusM17NEngineClass    *klass);
-static void	ibus_m17n_engine_init		    (IBusM17NEngine		    *m17n);
+static void ibus_m17n_engine_class_init     (IBusM17NEngineClass    *klass);
+static void ibus_m17n_engine_init           (IBusM17NEngine         *m17n);
 static GObject*
             ibus_m17n_engine_constructor    (GType                   type,
                                              guint                   n_construct_params,
                                              GObjectConstructParam  *construct_params);
-static void	ibus_m17n_engine_destroy		(IBusM17NEngine		    *m17n);
+static void ibus_m17n_engine_destroy        (IBusM17NEngine         *m17n);
 static gboolean
-			ibus_m17n_engine_process_key_event
+            ibus_m17n_engine_process_key_event
                                             (IBusEngine             *engine,
-                                             guint               	 keyval,
-                                             guint               	 modifiers);
+                                             guint                   keyval,
+                                             guint                   keycode,
+                                             guint                   modifiers);
 static void ibus_m17n_engine_focus_in       (IBusEngine             *engine);
 static void ibus_m17n_engine_focus_out      (IBusEngine             *engine);
 static void ibus_m17n_engine_reset          (IBusEngine             *engine);
@@ -53,14 +54,15 @@ static void ibus_m17n_engine_page_up        (IBusEngine             *engine);
 static void ibus_m17n_engine_page_down      (IBusEngine             *engine);
 static void ibus_m17n_engine_cursor_up      (IBusEngine             *engine);
 static void ibus_m17n_engine_cursor_down    (IBusEngine             *engine);
-static void ibus_m17n_property_activate     (IBusEngine             *engine,
+static void ibus_m17n_engine_property_activate
+                                            (IBusEngine             *engine,
                                              const gchar            *prop_name,
-                                             gint                    prop_state);
+                                             guint                   prop_state);
 static void ibus_m17n_engine_property_show
-											(IBusEngine             *engine,
+                                            (IBusEngine             *engine,
                                              const gchar            *prop_name);
 static void ibus_m17n_engine_property_hide
-											(IBusEngine             *engine,
+                                            (IBusEngine             *engine,
                                              const gchar            *prop_name);
 
 static void ibus_m17n_engine_commit_string
@@ -75,41 +77,41 @@ static GHashTable      *im_table = NULL;
 GType
 ibus_m17n_engine_get_type (void)
 {
-	static GType type = 0;
+    static GType type = 0;
 
-	static const GTypeInfo type_info = {
-		sizeof (IBusM17NEngineClass),
-		(GBaseInitFunc)		NULL,
-		(GBaseFinalizeFunc) NULL,
-		(GClassInitFunc)	ibus_m17n_engine_class_init,
-		NULL,
-		NULL,
-		sizeof (IBusM17NEngine),
-		0,
-		(GInstanceInitFunc)	ibus_m17n_engine_init,
-	};
+    static const GTypeInfo type_info = {
+        sizeof (IBusM17NEngineClass),
+        (GBaseInitFunc)        NULL,
+        (GBaseFinalizeFunc) NULL,
+        (GClassInitFunc)    ibus_m17n_engine_class_init,
+        NULL,
+        NULL,
+        sizeof (IBusM17NEngine),
+        0,
+        (GInstanceInitFunc)    ibus_m17n_engine_init,
+    };
 
-	if (type == 0) {
-		type = g_type_register_static (IBUS_TYPE_ENGINE,
-									   "IBusM17NEngine",
-									   &type_info,
-									   (GTypeFlags) 0);
-	}
+    if (type == 0) {
+        type = g_type_register_static (IBUS_TYPE_ENGINE,
+                                       "IBusM17NEngine",
+                                       &type_info,
+                                       (GTypeFlags) 0);
+    }
 
-	return type;
+    return type;
 }
 
 static void
 ibus_m17n_engine_class_init (IBusM17NEngineClass *klass)
 {
     GObjectClass *object_class = G_OBJECT_CLASS (klass);
-	IBusObjectClass *ibus_object_class = IBUS_OBJECT_CLASS (klass);
-	IBusEngineClass *engine_class = IBUS_ENGINE_CLASS (klass);
+    IBusObjectClass *ibus_object_class = IBUS_OBJECT_CLASS (klass);
+    IBusEngineClass *engine_class = IBUS_ENGINE_CLASS (klass);
 
-	parent_class = (IBusEngineClass *) g_type_class_peek_parent (klass);
+    parent_class = (IBusEngineClass *) g_type_class_peek_parent (klass);
 
     object_class->constructor = ibus_m17n_engine_constructor;
-	ibus_object_class->destroy = (IBusObjectDestroyFunc) ibus_m17n_engine_destroy;
+    ibus_object_class->destroy = (IBusObjectDestroyFunc) ibus_m17n_engine_destroy;
 
     engine_class->process_key_event = ibus_m17n_engine_process_key_event;
 
@@ -126,6 +128,7 @@ ibus_m17n_engine_class_init (IBusM17NEngineClass *klass)
     engine_class->cursor_up = ibus_m17n_engine_cursor_up;
     engine_class->cursor_down = ibus_m17n_engine_cursor_down;
 
+    engine_class->property_activate = ibus_m17n_engine_property_activate;
 }
 
 static void
@@ -171,7 +174,7 @@ ibus_m17n_engine_constructor (GType                   type,
                                           (GDestroyNotify) minput_close_im);
     }
 
-    im = g_hash_table_lookup (im_table, engine_name);
+    im = (MInputMethod *) g_hash_table_lookup (im_table, engine_name);
     if (im == NULL) {
         gchar *lang;
         gchar *name;
@@ -216,24 +219,7 @@ ibus_m17n_engine_constructor (GType                   type,
     m17n->context = minput_create_ic (im, NULL);
     mplist_add (m17n->context->plist, msymbol ("IBusEngine"), m17n);
 
-    /*
-        Minput_preedit_start
-        Minput_preedit_draw
-        Minput_preedit_done
-        Minput_status_start
-        Minput_status_draw
-        Minput_status_done
-        Minput_candidates_start
-        Minput_candidates_draw
-        Minput_candidates_done
-        Minput_set_spot
-        Minput_toggle
-        Minput_reset
-        Minput_get_surrounding_text
-        Minput_delete_surrounding_text
-    */
-
-    return (GObject *)m17n;
+    return (GObject *) m17n;
 }
 
 
@@ -260,7 +246,7 @@ ibus_m17n_engine_destroy (IBusM17NEngine *m17n)
         m17n->context = NULL;
     }
 
-	IBUS_OBJECT_CLASS (parent_class)->destroy ((IBusObject *)m17n);
+    IBUS_OBJECT_CLASS (parent_class)->destroy ((IBusObject *)m17n);
 }
 
 static void
@@ -277,47 +263,67 @@ MSymbol
 ibus_m17n_key_event_to_symbol (guint keyval,
                                guint modifiers)
 {
-    GString *key;
-    MSymbol mkey = Mnil;
+    GString *keysym;
+    MSymbol mkeysym = Mnil;
     guint mask = 0;
 
+    if (keyval >= IBUS_Shift_L && keyval <= IBUS_Hyper_R) {
+        return Mnil;
+    }
+
+    keysym = g_string_new ("");
+
     if (keyval >= IBUS_space && keyval <= IBUS_asciitilde) {
+        gint c = keyval;
         if (keyval == IBUS_space && modifiers & IBUS_SHIFT_MASK)
             mask |= IBUS_SHIFT_MASK;
 
         if (modifiers & IBUS_CONTROL_MASK) {
-            if (keyval >= IBUS_a && keyval <= IBUS_z)
-                keyval += IBUS_A - IBUS_a;
+            if (c >= IBUS_a && c <= IBUS_z)
+                c += IBUS_A - IBUS_a;
             mask |= IBUS_CONTROL_MASK;
         }
+
+        g_string_append_c (keysym, c);
     }
-    else if (keyval >= IBUS_Shift_L && keyval <= IBUS_Hyper_R) {
-        return Mnil;
+    else {
+        mask |= modifiers & (IBUS_CONTROL_MASK | IBUS_SHIFT_MASK);
+        g_string_append (keysym, ibus_keyval_name (keyval));
+        if (keysym->len == 0) {
+            g_string_free (keysym, TRUE);
+            return Mnil;
+        }
     }
 
-    mask |= modifiers & (IBUS_MOD1_MASK | IBUS_META_MASK);
+    mask |= modifiers & (IBUS_MOD1_MASK |
+                         IBUS_META_MASK |
+                         IBUS_SUPER_MASK |
+                         IBUS_HYPER_MASK);
 
-    key = g_string_new ("");
 
+    if (mask & IBUS_HYPER_MASK) {
+        g_string_prepend (keysym, "H-");
+    }
+    if (mask & IBUS_SUPER_MASK) {
+        g_string_prepend (keysym, "s-");
+    }
     if (mask & IBUS_MOD1_MASK) {
-        g_string_append (key, "A-");
+        g_string_prepend (keysym, "A-");
     }
     if (mask & IBUS_META_MASK) {
-        g_string_append (key, "M-");
+        g_string_prepend (keysym, "M-");
     }
     if (mask & IBUS_CONTROL_MASK) {
-        g_string_append (key, "C-");
+        g_string_prepend (keysym, "C-");
     }
     if (mask & IBUS_SHIFT_MASK) {
-        g_string_append (key, "S-");
+        g_string_prepend (keysym, "S-");
     }
 
-    g_string_append (key, ibus_keyval_name (keyval));
+    mkeysym = msymbol (keysym->str);
+    g_string_free (keysym, TRUE);
 
-    mkey = msymbol (key->str);
-    g_string_free (key, TRUE);
-
-    return mkey;
+    return mkeysym;
 }
 
 static gboolean
@@ -356,6 +362,7 @@ ibus_m17n_engine_process_key (IBusM17NEngine *m17n,
 static gboolean
 ibus_m17n_engine_process_key_event (IBusEngine     *engine,
                                     guint           keyval,
+                                    guint           keycode,
                                     guint           modifiers)
 {
     IBusM17NEngine *m17n = (IBusM17NEngine *) engine;
@@ -456,6 +463,14 @@ ibus_m17n_engine_cursor_down (IBusEngine *engine)
     parent_class->cursor_down (engine);
 }
 
+static void
+ibus_m17n_engine_property_activate (IBusEngine  *engine,
+                                    const gchar *prop_name,
+                                    guint        prop_state)
+{
+    parent_class->property_activate (engine, prop_name, prop_state);
+}
+
 static void
 ibus_m17n_engine_update_lookup_table (IBusM17NEngine *m17n)
 {
@@ -503,10 +518,15 @@ ibus_m17n_engine_update_lookup_table (IBusM17NEngine *m17n)
             ibus_lookup_table_set_page_size (m17n->table, mplist_length (p));
 
             for (; mplist_key (p) != Mnil; p = mplist_next (p)) {
-                gchar * buf;
-                buf = ibus_m17n_mtext_to_utf8 ((MText *) mplist_value (p));
-                ibus_lookup_table_append_candidate (m17n->table, ibus_text_new_from_string (buf));
-                g_free (buf);
+                MText *mtext;
+                gchar *buf;
+
+                mtext = (MText *) mplist_value (p);
+                buf = ibus_m17n_mtext_to_utf8 (mtext);
+                if (buf) {
+                    ibus_lookup_table_append_candidate (m17n->table, ibus_text_new_from_string (buf));
+                    g_free (buf);
+                }
             }
         }
 
@@ -529,15 +549,15 @@ static void
 ibus_m17n_engine_callback (MInputContext *context,
                            MSymbol        command)
 {
-    IBusM17NEngine *m17n;
-    MPlist *p;
+    IBusM17NEngine *m17n = NULL;
+    MPlist *p = NULL;
 
     p = mplist_find_by_key (context->plist,  msymbol ("IBusEngine"));
     if (p) {
-        m17n = (IBusM17NEngine *)mplist_value (p);
+        m17n = (IBusM17NEngine *) mplist_value (p);
     }
-    else {
-        m17n = NULL;
+
+    if (m17n == NULL) {
         return;
     }
 
@@ -547,18 +567,19 @@ ibus_m17n_engine_callback (MInputContext *context,
     else if (command == Minput_preedit_draw) {
         IBusText *text;
         gchar *buf;
-        gint len;
 
         buf = ibus_m17n_mtext_to_utf8 (m17n->context->preedit);
-        text = ibus_text_new_from_static_string (buf);
-        ibus_text_append_attribute (text, IBUS_ATTR_TYPE_FOREGROUND, 0x00ffffff, 0, -1);
-        ibus_text_append_attribute (text, IBUS_ATTR_TYPE_BACKGROUND, 0x00000000, 0, -1);
-        ibus_engine_update_preedit_text ((IBusEngine *)m17n,
-                                         text,
-                                         m17n->context->cursor_pos,
-                                         mtext_len (m17n->context->preedit) > 0);
-        g_object_unref (text);
-        g_free (buf);
+        if (buf) {
+            text = ibus_text_new_from_string (buf);
+            ibus_text_append_attribute (text, IBUS_ATTR_TYPE_FOREGROUND, 0x00ffffff, 0, -1);
+            ibus_text_append_attribute (text, IBUS_ATTR_TYPE_BACKGROUND, 0x00000000, 0, -1);
+            ibus_engine_update_preedit_text ((IBusEngine *) m17n,
+                                             text,
+                                             m17n->context->cursor_pos,
+                                             mtext_len (m17n->context->preedit) > 0);
+            g_object_unref (text);
+            g_free (buf);
+        }
     }
     else if (command == Minput_preedit_done) {
         ibus_engine_hide_preedit_text ((IBusEngine *)m17n);
@@ -581,6 +602,7 @@ ibus_m17n_engine_callback (MInputContext *context,
             ibus_property_set_label (m17n->status_prop, NULL);
             ibus_property_set_visible (m17n->status_prop, FALSE);
         }
+
         ibus_engine_update_property ((IBusEngine *)m17n, m17n->status_prop);
         g_free (status);
     }
